"""Get data from psql"""

# flake8: noqa E501

# import asyncio
import json
import os

import psycopg
import psycopg_infdate
import yaml
from icecream import ic
from psycopg.rows import dict_row

from dump_psql_roles_grants.commands import commands as cmd

# DEBUG_DB_INCLUDE_ONLY = ["dev-wallet"]
# DEBUG_DB_NUM = 2
DEBUG_DB_INCLUDE_ONLY = []
DEBUG_DB_NUM = None

result = {}

psycopg_infdate.register_inf_date_handler(psycopg)


def get_something(cur, sql, ignore_columns, ignore_rows=None):
    """Common func to get something from psql"""

    something = {}
    cur.execute(sql)
    for row in cur:
        name = row[ignore_columns[0]]
        for ignore_column in ignore_columns:
            del row[ignore_column]
        something[name] = dict(row)
    if ignore_rows is not None:
        for r in ignore_rows:
            if r in something:
                del something[r]
    return something


def get_tables_info(
    input_, instance, dbname, user, host, password, ignores
):  # pylint: disable=too-many-arguments, too-many-locals
    """Get info about tables: schemas, owners, perms"""

    def get_schemas(cur):
        ignore_columns = ["Name"]
        return get_something(cur, cmd.get_schemas, ignore_columns, ignores.schema)

    def get_owners(cur, schema):
        cmddt = (cmd.dt).replace("PLACEHOLDER", schema)
        ignore_columns = ["Name", "Type", "Schema"]
        return get_something(cur, cmddt, ignore_columns)

    def get_tables_perms(cur, schema):
        cmdz = (cmd.z).replace("PLACEHOLDER", schema)
        ignore_columns = ["Name", "Schema"]
        return get_something(cur, cmdz, ignore_columns, ignores.table)

    tables_info = {}
    conn_str = f"{dbname=} {user=} {host=} {password=} connect_timeout=3"
    # with psycopg.AsyncConnection.connect(
    with psycopg.connect(  # pylint: disable=not-context-manager
        conn_str, row_factory=dict_row
    ) as conn:
        with conn.cursor() as cur:
            schemas = get_schemas(cur)
            ic(input_, instance, dbname, list(schemas))
            for schema in schemas:
                ic(input_, instance, dbname, schema)

                owners = get_owners(cur, schema)
                owners_empty = not bool(owners)
                ic(input_, instance, dbname, schema, owners_empty)

                tables_perm = get_tables_perms(cur, schema)
                tables_perm_empty = not bool(tables_perm)
                ic(input_, instance, dbname, schema, tables_perm_empty)

                # merge owners to tables_perm
                for table, vals in owners.items():
                    tables_perm[table] |= vals

                # drop not type=table items, drop Type proper
                for table in list(tables_perm):
                    type_ = tables_perm[table]["Type"]
                    del tables_perm[table]["Type"]
                    if type_ != "table":
                        del tables_perm[table]

                if tables_perm:
                    tables_info[schema] = tables_perm
    result[input_][instance]["databases"][dbname]["Schemas"] = tables_info


def get_instance_info(
    what, input_, instance, ignores, dbname, user, host, password
):  # pylint: disable=too-many-arguments, too-many-locals
    """Get info about instance: databases and roles"""

    def get_databases(cur):
        return get_something(cur, cmd.l, ["Name"], ignores.db)

    def get_roles(cur):
        return get_something(cur, cmd.du, ["rolname"], ignores.role)

    instance_info = {}
    conn_str = f"{dbname=} {user=} {host=} {password=} connect_timeout=3"
    ic(input_, instance, conn_str)
    # with psycopg.AsyncConnection.connect(
    with psycopg.connect(  # pylint: disable=not-context-manager
        conn_str, row_factory=dict_row
    ) as conn:
        with conn.cursor() as cur:
            if what != "roles":
                databases = get_databases(cur)
                if DEBUG_DB_INCLUDE_ONLY:
                    for db in list(databases):
                        if db not in DEBUG_DB_INCLUDE_ONLY:
                            del databases[db]
                instance_info["databases"] = databases
            if what != "databases":
                instance_info["roles"] = get_roles(cur)

    result[input_][instance] = instance_info


def process_instance(
    input_, instance, conn_info, what, count_str_instance, ignores
):  # pylint: disable=too-many-arguments
    """Get data for instance"""

    get_instance_info(what, input_, instance, ignores, **conn_info)

    if what != "roles":
        databases = result[input_][instance]["databases"]
        databases_number = len(databases)
        ic(databases_number)

        # block for useless log output
        for count, dbname in enumerate(list(databases)[:DEBUG_DB_NUM]):
            if DEBUG_DB_INCLUDE_ONLY and dbname not in DEBUG_DB_INCLUDE_ONLY:
                continue

            count_str_db = (
                f"{count+1:{0}{len(str(databases_number))}}/{databases_number}"
            )
            ic(input_, instance, count_str_instance, dbname, count_str_db)

            get_tables_info(
                **conn_info
                | {
                    "input_": input_,
                    "instance": instance,
                    "dbname": dbname,
                    "ignores": ignores,
                }
            )

        # asyncio.gather(
        #     *(
        #         get_tables_info(
        #             **conn_info
        #             | {"input_": input_, "instance": instance, "dbname": dbname}
        #         )
        #         for dbname in list(databases)[:DEBUG_DB_NUM]
        #     )
        # )


def process_input_(input_, instances, what, ignores):
    """Get data for input"""

    # add root user to ignore roles list
    # for conn_info in instances.values():
    #     ignores.role.append(conn_info["user"])

    # asyncio.gather(
    #     *(
    #         process_instance(input_, instance, conn_info, what)
    #         for instance, conn_info in instances.items()
    #     )
    # )
    instances_number = len(instances.keys())
    ic(input_, instances_number)
    for count, (instance, conn_info) in enumerate(instances.items()):
        count_str_instance = (
            f"{count+1:{0}{len(str(instances_number))}}/{instances_number}"
        )
        process_instance(input_, instance, conn_info, what, count_str_instance, ignores)


def get_instances(file_name, inputs_dir):
    """Get data from input files"""

    with open(os.path.join(inputs_dir, file_name), encoding="utf8") as f:
        ext = os.path.splitext(file_name)[1]
        if ext == ".json":
            instances = json.load(f)
        elif ext == ".yaml":
            instances = yaml.safe_load(f)
        else:
            raise IOError(f"Not supported format for '{file_name}'")
    return instances


def get_data_from_psql(what, inputs_dir, ignores):
    """Get databases and roles info main func"""

    inputs_dict = {}
    for file_name in os.listdir(inputs_dir):
        input_ = os.path.splitext(file_name)[0]
        if input_ in ignores.inputs:
            continue
        inputs_dict[input_] = get_instances(file_name, inputs_dir)

    # asyncio.gather(
    #     *(
    #         process_input_(input_, instances, what)
    #         for input_, instances in inputs_dict.items()
    #     )
    # )

    for input_, instances in inputs_dict.items():
        result[input_] = {}
        process_input_(input_, instances, what, ignores)
    return ic(result)


# def get_data_from_psql(what):
#     """Fire for async_get_data_from_psql func"""

#     asyncio.run(async_get_data_from_psql(what))
#     ic(result)
#     return result



#!/usr/bin/env python3
"""Print and dump results in table view"""

import os

import pytablewriter
from icecream import ic

from dump_psql_roles_grants.config import HEADER_DB, HEADER_ROLES


def get_table_formats():
    """Get supported TableWriters in pytablewriter"""

    formats = []
    for elem in dir(pytablewriter):
        if elem.endswith("TableWriter"):
            formats.append(elem.rstrip("TableWriter"))
    return formats


def fill_ignored_indices(header, ignored_cols):
    """Convert ignored header names into indices"""

    ignored_indices = []
    for ignored_col in ignored_cols:
        index = header.index(ignored_col)
        ignored_indices.append(index)
    return ignored_indices


def table_filter(header, ignored_indices, rows):
    """Remove ignored columns from table"""

    f_header = [x for x in header if header.index(x) not in ignored_indices]
    f_rows = []
    for row in rows:
        for col in sorted(ignored_indices, reverse=True):
            _ = row.pop(col)
        f_rows.append(row)
    return f_header, f_rows


def table_gen_roles(roles, rows, row, _):
    """Role specific table gen"""

    role_column = len(row)
    for role, role_props in roles.items():
        row.append(role)
        row.append(role_props["rolbypassrls"])
        row.append(role_props["rolcanlogin"])
        row.append(role_props["rolconnlimit"])
        row.append(role_props["rolcreatedb"])
        row.append(role_props["rolcreaterole"])
        row.append(role_props["rolinherit"])
        row.append(role_props["rolreplication"])
        row.append(role_props["rolsuper"])
        row.append(role_props["rolvaliduntil"])
        row.append("\n".join(role_props["memberof"]))
        rows.append(row)
        row = row[:role_column]


def table_gen_dbs(databases, rows, row, header):
    """Database specific table gen"""

    instance_name_column = len(row)
    for db, db_props in databases.items():

        row.append(db)
        row.append(db_props["Owner"])
        row.append(db_props["Access privileges"])

        access_privileges_column = len(row)
        if db_props["Schemas"]:
            for schema, tables in db_props["Schemas"].items():
                row.append(schema)

                schema_column = len(row)
                for table, table_props in tables.items():
                    row.append(table)
                    row.append(table_props.get("Owner"))
                    row.append(table_props.get("Access privileges"))
                    row.append(table_props.get("Column privileges"))
                    row.append(table_props.get("Policies"))

                    rows.append(row)
                    row = row[:schema_column]
                row = row[: schema_column - 1]
        else:
            row += [""] * (len(header) - len(row))
            rows.append(row)
            row = row[: access_privileges_column - 3]
        row = row[:instance_name_column]


def table_gen(conts, entity, entity_func, header, ignored_cols):
    """Generate common part of a table"""

    ignored_indices = fill_ignored_indices(header, ignored_cols)

    rows = []
    for input_, instances in conts.items():
        ic(input_, len(instances))
        for instance, entities in instances.items():
            row = []
            row.append(input_)
            row.append(instance)
            entity_func(entities[entity], rows, row, header)

    header, rows = table_filter(header, ignored_indices, rows)
    # print(len(header), header)
    # for row in rows:
    #     print(len(row), row)
    # print()
    return header, rows


def table_write(
    table_name,
    headers,
    value_matrix,
    print_=False,
    dump=False,
    fmt="Markdown",
    results_dir=".",
):  # pylint: disable=too-many-arguments
    """Print and dump results in table view implementation"""

    writer = getattr(pytablewriter, f"{fmt}TableWriter")(
        table_name=table_name.capitalize(),
        headers=headers,
        value_matrix=value_matrix,
        margin=1,
    )
    if print_:
        writer.write_table()
    if dump:
        os.makedirs(results_dir, exist_ok=True)
        results_file = f"{table_name}.{fmt.lower()}"
        writer.dump(os.path.join(results_dir, results_file))


def dump_table(
    print_, what, data, fmt, results_dir, ignores
):  # pylint: disable=too-many-arguments
    """Print and dump results in table view main func"""

    entities = {
        "databases": {
            "entity_func": table_gen_dbs,
            "header": HEADER_DB,
            "ignored_cols": ignores.result_table_column_databases,
        },
        "roles": {
            "entity_func": table_gen_roles,
            "header": HEADER_ROLES,
            "ignored_cols": ignores.result_table_column_roles,
        },
    }

    if what == "roles":
        del entities["databases"]
    elif what == "databases":
        del entities["roles"]

    ic(bool(data))
    for entity, settings in entities.items():
        headers, rows = table_gen(**(settings | {"conts": data, "entity": entity}))
        if print_:
            table_write(
                entity,
                headers,
                rows,
                print_=True,
                fmt="Markdown",
                dump=False,
                results_dir=results_dir,
            )
        if fmt:
            table_write(
                entity,
                headers,
                rows,
                print_=False,
                fmt=fmt,
                dump=True,
                results_dir=results_dir,
            )



"""Configuration-related things"""

import yaml

HEADER_DB = [
    "Env",
    "Instance name",
    "Database",
    "Owner",
    "Access privileges",
    "Schema",
    "Table",
    "Table owner",
    "Access privileges",
    "Column privileges",
    "Policies",
]

HEADER_ROLES = [
    "Env",
    "Instance name",
    "Role name",
    "rolbypassrls",
    "rolcanlogin",
    "rolconnlimit",
    "rolcreatedb",
    "rolcreaterole",
    "rolinherit",
    "rolreplication",
    "rolsuper",
    "rolvaliduntil",
    "memberof",
]


def load_config(conf_file):
    """Load config from file"""

    try:
        with open(conf_file, "r", encoding="utf8") as f:
            conf = yaml.safe_load(f)
    except FileNotFoundError as exc:
        raise SystemExit(
            f"Config file {conf_file} not found. "
            "You can create it with `dump_psql_roles_grants gen-example-config`"
        ) from exc
    return conf



# pylint: disable=missing-module-docstring
# flake8: noqa E501

from dump_psql_roles_grants.classes import Commands

commands = Commands(
    # \l
    l="""
    SELECT
      d.datname as "Name",
      pg_catalog.pg_get_userbyid(d.datdba) as "Owner",
      pg_catalog.array_to_string(d.datacl, E'\n') AS "Access privileges"
    FROM pg_catalog.pg_database d
    ORDER BY 1;
    """,
    # \du
    du="""
    SELECT r.rolname, r.rolsuper, r.rolinherit,
      r.rolcreaterole, r.rolcreatedb, r.rolcanlogin,
      r.rolconnlimit, r.rolvaliduntil,
      ARRAY(SELECT b.rolname
            FROM pg_catalog.pg_auth_members m
            JOIN pg_catalog.pg_roles b ON (m.roleid = b.oid)
            WHERE m.member = r.oid) as memberof
    , r.rolreplication
    , r.rolbypassrls
    FROM pg_catalog.pg_roles r
    WHERE r.rolname !~ '^pg_'
    ORDER BY 1;
    """,
    dt="""
    SELECT n.nspname as "Schema",
      c.relname as "Name",
      CASE c.relkind WHEN 'r' THEN 'table' WHEN 'v' THEN 'view' WHEN 'm' THEN 'materialized view' WHEN 'i' THEN 'index' WHEN 'S' THEN 'sequence' WHEN 's' THEN 'special' WHEN 't' THEN 'TOAST table' WHEN 'f' THEN 'foreign table' WHEN 'p' THEN 'partitioned table' WHEN 'I' THEN 'partitioned index' END as "Type",
      pg_catalog.pg_get_userbyid(c.relowner) as "Owner"
    FROM pg_catalog.pg_class c
         LEFT JOIN pg_catalog.pg_namespace n ON n.oid = c.relnamespace
         LEFT JOIN pg_catalog.pg_am am ON am.oid = c.relam
    WHERE c.relkind IN ('r','p','t','s','')
      AND n.nspname OPERATOR(pg_catalog.~) '^(PLACEHOLDER)$' COLLATE pg_catalog.default
    ORDER BY 1,2;
    """,
    get_schemas="""
    SELECT schema_name as "Name"
    FROM information_schema.schemata;
    """,
    z="""
    SELECT n.nspname as "Schema",
      c.relname as "Name",
      CASE c.relkind WHEN 'r' THEN 'table' WHEN 'v' THEN 'view' WHEN 'm' THEN 'materialized view' WHEN 'S' THEN 'sequence' WHEN 'f' THEN 'foreign table' WHEN 'p' THEN 'partitioned table' END as "Type",
      pg_catalog.array_to_string(c.relacl, E'\n') AS "Access privileges",
      pg_catalog.array_to_string(ARRAY(
        SELECT attname || E':\n  ' || pg_catalog.array_to_string(attacl, E'\n  ')
        FROM pg_catalog.pg_attribute a
        WHERE attrelid = c.oid AND NOT attisdropped AND attacl IS NOT NULL
      ), E'\n') AS "Column privileges",
      pg_catalog.array_to_string(ARRAY(
        SELECT polname
        || CASE WHEN NOT polpermissive THEN
           E' (RESTRICTIVE)'
           ELSE '' END
        || CASE WHEN polcmd != '*' THEN
               E' (' || polcmd::pg_catalog.text || E'):'
           ELSE E':'
           END
        || CASE WHEN polqual IS NOT NULL THEN
               E'\n  (u): ' || pg_catalog.pg_get_expr(polqual, polrelid)
           ELSE E''
           END
        || CASE WHEN polwithcheck IS NOT NULL THEN
               E'\n  (c): ' || pg_catalog.pg_get_expr(polwithcheck, polrelid)
           ELSE E''
           END    || CASE WHEN polroles <> '{0}' THEN
               E'\n  to: ' || pg_catalog.array_to_string(
                   ARRAY(
                       SELECT rolname
                       FROM pg_catalog.pg_roles
                       WHERE oid = ANY (polroles)
                       ORDER BY 1
                   ), E', ')
           ELSE E''
           END
        FROM pg_catalog.pg_policy pol
        WHERE polrelid = c.oid), E'\n')
        AS "Policies"
    FROM pg_catalog.pg_class c
         LEFT JOIN pg_catalog.pg_namespace n ON n.oid = c.relnamespace
    WHERE c.relkind IN ('r','v','m','S','f','p')
      AND n.nspname OPERATOR(pg_catalog.~) '^(PLACEHOLDER)$' COLLATE pg_catalog.default
    ORDER BY 1, 2;
    """,
)



"""
This module serves as the entry point for the dump_psql_roles_grants application.

It imports the `cli` function from the `dump_psql_roles_grants.cli` module and calls
it when the script is executed.
"""

from dump_psql_roles_grants.cli import cli

if __name__ == "__main__":
    cli()



#!/usr/bin/env python
"""Command line interface"""

import click

from dump_psql_roles_grants.classes import Ignores
from dump_psql_roles_grants.config import load_config
from dump_psql_roles_grants.get_data import get_data_from_psql
from dump_psql_roles_grants.table import dump_table

# from table import get_table_formats


@click.group()
# @click.option("-v", "--verbose", count=True)
def cli():
    """Dump Postgres Roles and Grants"""


@cli.command()
@click.option(
    "--print/--no-print",
    "print_",
    is_flag=True,
    show_default=True,
    default=True,
    help="display results",
)
@click.option(
    "--what",
    "-w",
    default="all",
    type=click.Choice(["all", "databases", "roles"]),
    show_default=True,
    help="what to export",
)
@click.option(
    "--dump-in-format",
    "-f",
    default="Csv",
    # type=click.Choice(get_table_formats()),
    type=click.Choice(
        [
            "Abstrac",
            "AsciiDoc",
            "BoldUnicod",
            "Borderless",
            "Css",
            "Csv",
            "ExcelXls",
            "ExcelXlsx",
            "Htm",
            "JavaScrip",
            "JsonLines",
            "Json",
            "Latex",
            "Ltsv",
            "Markdown",
        ]
    ),
    help="dump results in specified table format",
)
@click.option(
    "--config",
    "-c",
    default="config.yaml",
    show_default=True,
    help="config file to use",
)
def export(print_, what, dump_in_format, config):
    """Export instance data"""

    _config = load_config(config)
    ignores = Ignores(**_config["ignores"])
    inputs_dir = _config["inputs_dir"]
    results_dir = _config["results_dir"]

    data = get_data_from_psql(what, inputs_dir, ignores)

    dump_table(print_, what, data, dump_in_format, results_dir, ignores)


@cli.command()
@click.option(
    "--output",
    "-o",
    show_default=True,
    help="file to write example config to",
)
def gen_example_config(output=None):
    """Generate example config file"""

    example_config = """
# Example config file

inputs_dir: ./_inputs
results_dir: ./_results
ignores:
  inputs:
    # []
    - dev
    # - stage
    - test
    - prod
  db:
    - postgres
    - rdsadmin
    - template0
    - template1
  role:
    - rdstopmgr
    - rdsrepladmin
    - rdsadmin
    - rds_superuser
    - rds_replication
    - rds_password
    - rds_iam
    - rds_ad
    - datadog_agent
    - change_owner
  schema:
    - pg_catalog
    - information_schema
    - datadog
  table:
    - pg_stat_statements
  result_table_column_databases:
    []
    # - Env
    # - Instance name
    # - Database
    # - Owner
    # - Access privileges
    # - Schema
    # - Table
    # - Table owner
    # - Access privileges
    # - Column privileges
    # - Policies
  result_table_column_roles:
    []
    # - Env
    # - Instance name
    # - Role name
    # - rolbypassrls
    # - rolcanlogin
    # - rolconnlimit
    # - rolcreatedb
    # - rolcreaterole
    # - rolinherit
    # - rolreplication
    # - rolsuper
    # - rolvaliduntil
    # - memberof
""".lstrip()

    if output:
        with open(output, "w", encoding="utf-8") as f:
            f.write(example_config)
    else:
        click.echo(example_config)



"""Dataclasses for app to be able to access to props by dot"""

from dataclasses import dataclass


@dataclass
class Commands:
    """Sql commands spec"""

    dt: str
    du: str
    get_schemas: str
    l: str
    z: str


@dataclass
class Ignores:
    """Ignores spec"""

    db: list
    inputs: list
    role: list
    schema: list
    table: list
    result_table_column_databases: list
    result_table_column_roles: list
